#include "cylinderEnvironment.hh"
#include <cmath>
#include "mtrand.h"
#include <ode/ode.h>
#include <vtkActor.h>
#include <vtkCylinderSource.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include "molecule.hh"
#include "type.hh"
#include "vector.hh"

CylinderEnvironment::CylinderEnvironment(double height, double temperature) : height(height), Environment(temperature)
{
	src = vtkCylinderSource::New();
	src->SetCapping(1);
	src->SetHeight(height);
	src->SetRadius(RADIUS);
	src->SetResolution(THETA_RESOLUTION);

	map = vtkPolyDataMapper::New();
	map->SetInput(src->GetOutput());

	graphicRepresentation->SetMapper(map);

	this->createGeometry();
}

CylinderEnvironment::~CylinderEnvironment()
{
	int j;
	for (j = 0; j < THETA_RESOLUTION; j++)
		delete planes[j];
	delete top;
	delete bottom;
	map->Delete();
	src->Delete();
}

Vector CylinderEnvironment::generateRandomValidPosition(Molecule *m)
{
	MTRand drand;
	Vector tempPosition, direction;
	Vector unitY = Vector(0.0, 1.0, 0.0);
	double minY = -this->height / 2.0 + m->getType().getRadius();
	do {
		direction = Vector::randomDirection();
		direction = direction - (unitY * direction.getY());
		direction = direction.direction(); // normalize
		tempPosition = direction * drand() * (RADIUS - m->getType().getRadius());
		tempPosition = tempPosition + (unitY * (minY + drand() * (this->height - 2.0 * m->getType().getRadius())));
	} while (hasMoleculeNear(tempPosition, m->getType().getRadius()));
	return tempPosition;
}

double CylinderEnvironment::biggestFreePathInside() {
	return sqrt(RADIUS * RADIUS + this->height * this->height);
}

void CylinderEnvironment::createGeometry()
{
	double deltaTheta = 2.0 * M_PI / THETA_RESOLUTION;
	double theta;
	int iTheta;
	// adds the planes not in the poles; these two will be added later
	for (iTheta = 0, theta = 0.0; iTheta < THETA_RESOLUTION; iTheta++, theta += deltaTheta)
	{
		planes[iTheta] = new dPlane(space->id(), cos(theta), 0.0, sin(theta), -RADIUS);
	}
	top = new dPlane(space->id(), 0.0, -1.0, 0.0, -(height / 2.0));
	bottom = new dPlane(space->id(), 0.0, 1.0, 0.0, -(height / 2.0));
}