#include <vtkActor.h>
#include <vtkProperty.h>
#include "environment.hh"
#include "molecule.hh"
#include "mtrand.h"
#include "vector.hh"

#include <iostream>

Environment::Environment(double temperature)
{
	graphicRepresentation = vtkActor::New();
	graphicRepresentation->GetProperty()->SetOpacity(0.1);
	graphicRepresentation->GetProperty()->SetColor(1.0, 1.0, 1.0);
	this->temperature = temperature;
	physics = new dWorld();
	space = new dHashSpace(0);
	physics->setGravity(0.0, 0.0, 0.0); // no gravity
	molecules = std::map<dBodyID, Molecule *>();
}

Environment::~Environment()
{
	molecules.clear();
	graphicRepresentation->Delete();
	delete space;
	delete physics;
}

void Environment::updateMolecules()
{
	std::map<dBodyID, Molecule *>::iterator it;
	for (it = molecules.begin(); it != molecules.end(); it++)
		it->second->updatePhysics();
}

bool Environment::addMolecule(Molecule *m)
{
	std::pair<std::map<dBodyID, Molecule *>::iterator, bool> result;
	if (hasMolecule(m))
		return false;
	dBody *body = new dBody(physics->id());
	result = molecules.insert(std::pair<dBodyID, Molecule *>(body->id(), m));
	if (result.second)
	{
		dMass *mass = new dMass();
		mass->setSphere(1.0, m->getType().getRadius());
		mass->adjust(m->getType().getMass());
		body->setMass(mass);

		dGeom *geometry = new dSphere(space->id(), m->getType().getRadius());
		geometry->setBody(body->id());
		m->addPhysics(body);
	}
	else
		delete body;
	return result.second;
}

int Environment::getNumberOfMolecules()
{
	return molecules.size();
}

vtkActor *Environment::getGraphicRepresentation()
{
	return this->graphicRepresentation;
}

double Environment::getTemperature()
{
	return this->temperature;
}

void Environment::setTemperature(double temp)
{
	if (temp > 0.0)
		this->temperature = temp;
}

dWorld *Environment::getPhysics()
{
	return this->physics;
}

dSpace *Environment::getSpace()
{
	return this->space;
}

Vector Environment::generateRandomSpeedFor(Molecule *m)
{
	MTRand drand;
	double sum = 0.0;
	double maxSum = drand(); // where we will stop integrating
	double massKg = m->getType().getMass() * 1.0e-24; // mass of molecule in kg
	double c1 = 4.0 * M_PI *
		pow(massKg / (2.0 * M_PI * BOLTZMANN * temperature), 1.5);
	double c2 = -massKg / (2.0 * BOLTZMANN * temperature);
	double x = 0.0;
	while (sum < maxSum)
	{
		sum += (c1 * x * x * exp(c2 * x * x)) * DELTA;
		x += DELTA;
	}
	// now x contains a speed in m/s; convert it to Å/ns and use
	// a random direction
	x *= 10.0;
	return Vector::randomDirection() * x;
}

bool Environment::hasMoleculeNear(Vector position, double radius)
{
	std::map<dBodyID, Molecule *>::iterator it;
	for (it = molecules.begin(); it != molecules.end(); it++)
	{
		if ((it->second->getPosition() - position).length() <= it->second->getType().getRadius() + radius)
			return true;
	}
	return false;
}

double Environment::averageSpeedOfMolecules() {
	std::map<dBodyID, Molecule *>::iterator it;
	double sum = 0.0;
	for (it = molecules.begin(); it != molecules.end(); it++)
	{
		sum += it->second->getSpeed().length();
	}
	sum /= molecules.size();
	return sum;
}

double Environment::kineticEnergy()
{
	std::map<dBodyID, Molecule *>::iterator it;
	double total = 0.0;
	Molecule *m;
	for (it = molecules.begin(); it != molecules.end(); it++)
	{
		m = it->second;
		total += m->getType().getMass() * m->getSpeed().dot(m->getSpeed()) / 2.0;
	}
	return total;
}

Molecule *Environment::getMoleculeByBodyID(dBodyID id)
{
	std::map<dBodyID, Molecule *>::iterator result = molecules.find(id);
	if (result == molecules.end())
		return 0;
	return result->second;
}

std::map<dBodyID, Molecule *> Environment::getMolecules()
{
	return this->molecules;
}

bool Environment::hasMolecule(Molecule *m)
{
	std::map<dBodyID, Molecule *>::iterator it;
	for (it = molecules.begin(); it != molecules.end(); it++)
		if (it->second == m)
			return true;
	return false;
}