#include "molecule.hh"
#include <ode/ode.h>
#include <vtkActor.h>
#include <vtkDataSetMapper.h>
#include <vtkSphereSource.h>
#include <vtkPoints.h>
#include <vtkPolyData.h>
#include <vtkPolyDataMapper.h>
#include <vtkPolyLine.h>
#include <vtkProperty.h>
#include "configuration.hh"
#include "environment.hh"
#include "mtrand.h"
#include "path.hh"

Molecule::Molecule(Type t)
{
	type = t;
	lastSpeed = Vector();
	path = new Path();
	body = 0;
	initialPosition = 0;

	Configuration *conf = Configuration::getInstance();
	baseColor = conf->getMoleculesColor();
	collidedColor = conf->getCollidedMoleculesColor();
	colorFadeSpeed = conf->getColorChangeSpeed();

	src = vtkSphereSource::New();
	src->SetRadius(t.getRadius());
	src->SetPhiResolution(16);
	src->SetThetaResolution(16);

	map = vtkPolyDataMapper::New();
	map->SetInput(src->GetOutput());

	double rgba[4];
	conf->getMoleculesColor().copyToArray(rgba);
	graphicRepresentation = vtkActor::New();
	graphicRepresentation->SetMapper(map);
	graphicRepresentation->GetProperty()->SetColor(rgba);
}

Molecule::~Molecule()
{
	src->Delete();
	map->Delete();
	graphicRepresentation->Delete();
	delete path;
}

bool Molecule::addToEnvironment(Environment *env)
{
	if (env->addMolecule(this))
	{
		this->setPosition(env->generateRandomValidPosition(this));
		path->addPoint(this->getPosition());
		return true;
	}
	return false;
}

void Molecule::addPhysics(dBody *body)
{
	if (this->body != 0)
		delete body;
	this->body = body;
}

void Molecule::updatePhysics()
{
	const dReal *position = this->body->getPosition();
	Vector currentSpeed = getSpeed();
	this->graphicRepresentation->SetPosition(position[0], position[1], position[2]);
	if (shouldDrawPath)
	{
		if (lastSpeed.direction() != currentSpeed.direction())
			path->addPoint(Vector(position[0], position[1], position[2]));
		else
			path->updateEndPoint(Vector(position[0], position[1], position[2]));
	}

	Color c = this->getColor().fadeTo(this->baseColor, colorFadeSpeed);
	double rgba[4];
	c.copyToArray(rgba);
	this->graphicRepresentation->GetProperty()->SetColor(rgba);
	lastSpeed = currentSpeed;
}

void Molecule::collided()
{
	double rgba[4];
	collidedColor.copyToArray(rgba);
	this->graphicRepresentation->GetProperty()->SetColor(rgba);
}

Vector Molecule::getPosition()
{
	const dReal *position = this->body->getPosition();
	return Vector(position[0], position[1], position[2]);
}

void Molecule::setPosition(Vector p)
{
	if (initialPosition == 0)
		initialPosition = new Vector(p.getX(), p.getY(), p.getZ());
	this->graphicRepresentation->SetPosition(p.getX(), p.getY(), p.getZ());
	this->body->setPosition(p.getX(), p.getY(), p.getZ());
}

Vector Molecule::getInitialPosition()
{
	if (initialPosition == 0)
		return Vector();
	return *(this->initialPosition);
}

Vector Molecule::getSpeed()
{
	const dReal *speed = this->body->getLinearVel();
	return Vector(speed[0], speed[1], speed[2]);
}

void Molecule::setSpeed(Vector s)
{
	this->body->setLinearVel(s.getX(), s.getY(), s.getZ());
}

Color Molecule::getColor()
{
	double rgb[3];
	this->graphicRepresentation->GetProperty()->GetColor(rgb);
	return Color(rgb[0], rgb[1], rgb[2]);
}

void Molecule::setColor(Color c)
{
	this->baseColor = c;
	double rgba[4];
	c.copyToArray(rgba);
	graphicRepresentation->GetProperty()->SetColor(rgba);
}

void Molecule::setCollidedColor(Color c, double fadeSpeed)
{
	this->collidedColor = c;
	this->colorFadeSpeed = fadeSpeed;
}

void Molecule::drawPath(bool shouldDraw)
{
	shouldDrawPath = shouldDraw;
}

Type Molecule::getType()
{
	return this->type;
}

vtkActor *Molecule::getGraphicRepresentation()
{
	return this->graphicRepresentation;
}

Path *Molecule::getPath()
{
	return this->path;
}