#include "simulation.hh"
#include "collisionWatchers.hh"
#include "configuration.hh"
#include "environment.hh"
#include "environmentFactory.hh"
#include "molecule.hh"
#include "onScreenDiffusion.hh"
#include "path.hh"
#include "simulationVisualization.hh"

Simulation *Simulation::instance = 0;

Simulation *Simulation::getInstance()
{
	if (!instance)
		instance = new Simulation();
	return instance;
}

// ODE callback must be a global function
void callbackWrapper(void *data, dGeomID id1, dGeomID id2)
{
	((Simulation *) data)->treatCollisions(id1, id2);
}

Simulation::Simulation()
{
	this->contactJoints = new dJointGroup();
	this->env = 0;
	this->view = 0;
	this->atEnd = sigc::slot<void>();
	this->collisionWatchers = new CollisionWatchers();
}

bool Simulation::refresh()
{
	double interval, leftTime;
	if (elapsed >= maxDuration)
	{
		this->stop();
		this->atEnd();
		return true;
	}
	currentClock = clock();
	interval = (double) (currentClock - lastClock) / CLOCKS_PER_SEC;
	leftTime = interval;
	// strangely, sometimes the interval is zero...
	// in this case, we delay the update by doing nothing in this call
	while (leftTime >= STEP) {
		leftTime -= STEP;
		env->getSpace()->collide(this, &callbackWrapper);
		env->getPhysics()->step(STEP);
		contactJoints->empty();
	}
	if (interval >= STEP) {
		elapsed += (int) ((interval - leftTime) * 1000);
		env->updateMolecules();
		lastClock = currentClock;
	}
	if (this->view)
		this->view->update(interval);
	return true;
}

void Simulation::start()
{
	Configuration *conf = Configuration::getInstance();
	EnvironmentFactory *factory = new EnvironmentFactory();
	
	if (this->view)
		this->view->resetView();

	if (this->env)
		delete this->env;
	this->env = factory->createEnvironment(conf->getEnvironmentType(), conf->getEnvironmentScale(), conf->getTemperature());

	if (this->view)
		this->view->addVisibleItem(this->env->getGraphicRepresentation());
	
	this->addMoleculesToEnvironment();

	if (this->view && conf->shouldShowDiffusion())
		this->view->addOnScreenStat(new OnScreenDiffusion(this->env));

	this->elapsed = 0;
	this->maxDuration = conf->getDurationOfSimulation() * 1000.0; // we need it in ms
	this->unpause();
}

void Simulation::stop()
{
	signal.disconnect();
}

void Simulation::pause()
{
	signal.disconnect();
}

void Simulation::unpause()
{
	this->lastClock = clock();
	this->signal = Glib::signal_timeout().connect(sigc::mem_fun(*this, &Simulation::refresh), FRAME_DURATION);
}

void Simulation::treatCollisions(dGeomID id1, dGeomID id2)
{
	int i, contacts;
	dContact contactJoint[MAX_CONTACTS];
	dBodyID body1, body2;
	for (i = 0; i < MAX_CONTACTS; i++)
	{
		contactJoint[i].surface.mode = dContactBounce;
		contactJoint[i].surface.mu = 0; // No friction
		contactJoint[i].surface.bounce = 1.0 + 1.0/30.0 - 1.0/30000.0; // Collisions are totally elastic
		contactJoint[i].surface.bounce_vel = 0.0; // Minimum velocity to bounce
	}
	contacts = dCollide(id1, id2, MAX_CONTACTS, &contactJoint[0].geom, sizeof(dContact));
	if (contacts)
	{
		// Notify every CollisionWatcher that a collision ocurred
		this->collisionWatchers->notifyAll();
		body1 = dGeomGetBody(id1);
		body2 = dGeomGetBody(id2);
		// Collision between molecules; paint them with a different color for a while
		// if the user wants
		if (body1 && body2)
		{
			env->getMoleculeByBodyID(body1)->collided();
			env->getMoleculeByBodyID(body2)->collided();
		}
		for (i = 0; i < contacts; i++)
		{
			dJoint *joint = new dContactJoint(env->getPhysics()->id(), contactJoints->id(), &contactJoint[i]);
			joint->attach(body1, body2);
		}
	}
}

void Simulation::setVisualization(SimulationVisualization *view)
{
	this->view = view;
}

SimulationVisualization *Simulation::getVisualization()
{
	return this->view;
}

void Simulation::setFinishCallback(sigc::slot<void> atEnd)
{
	this->atEnd = atEnd;
}

CollisionWatchers *Simulation::getCollisionWatchers()
{
	return this->collisionWatchers;
}

void Simulation::addMoleculesToEnvironment()
{
	int i;
	Configuration *conf = Configuration::getInstance();
	Type type(conf->getRadiusOfMolecules());
	for (i = 0; i < conf->getNumberOfMolecules(); i++)
	{
		Molecule *m = new Molecule(type);
		m->addToEnvironment(this->env);
		m->setSpeed(this->env->generateRandomSpeedFor(m));
		m->getPath()->setColor(conf->getPathColor(i));
		m->setColor(conf->getMoleculesColor());
		m->setCollidedColor(conf->getCollidedMoleculesColor(), conf->getColorChangeSpeed());
		if (this->view)
			view->addVisibleItem(m->getGraphicRepresentation());
		if (conf->shouldDrawPaths())
		{
			m->drawPath(true);
			if (this->view)
				view->addVisibleItem(m->getPath()->getGraphics());
		}
		else
			m->drawPath(false);
	}
}